@@ -16,8 +16,13 @@
 # ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 # OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
+from typing import Callable, Dict
+
 from jinja2.ext import Extension
+from jinjafx import JinjaFx
+
 import netaddr, types
+
  
 class plugin(Extension):
   def __init__(self, environment):
@@ -137,7 +142,7 @@ def _cidr_lookup_query(v, iplist, value):
 def _first_usable_query(v, vtype):
     if vtype == "address":
         # Does it make sense to raise an error
-        raise Exception("Not a network address")
+        raise JinjaFx.TemplateError("Not a network address")
     elif vtype == "network":
         if v.size == 2:
             return str(netaddr.IPAddress(int(v.network)))
@@ -200,7 +205,7 @@ def _ipv6_query(v, value):
 def _last_usable_query(v, vtype):
     if vtype == "address":
         # Does it make sense to raise an error
-        raise Exception("Not a network address")
+        raise JinjaFx.TemplateError("Not a network address")
     elif vtype == "network":
         if v.size > 1:
             first_usable, last_usable = _first_last(v)
@@ -255,7 +260,7 @@ def _network_wildcard_query(v):
 def _next_usable_query(v, vtype):
     if vtype == "address":
         # Does it make sense to raise an error
-        raise Exception("Not a network address")
+        raise JinjaFx.TemplateError("Not a network address")
     elif vtype == "network":
         if v.size > 1:
             first_usable, last_usable = _first_last(v)
@@ -266,21 +271,21 @@ def _next_usable_query(v, vtype):
 
 def _peer_query(v, vtype):
     if vtype == "address":
-        raise Exception("Not a network address")
+        raise JinjaFx.TemplateError("Not a network address")
     elif vtype == "network":
         if v.size == 2:
             return str(netaddr.IPAddress(int(v.ip) ^ 1))
         if v.size == 4:
             if int(v.ip) % 4 == 0:
-                raise Exception(
+                raise JinjaFx.TemplateError(
                     "Network address of /30 has no peer"
                 )
             if int(v.ip) % 4 == 3:
-                raise Exception(
+                raise JinjaFx.TemplateError(
                     "Broadcast address of /30 has no peer"
                 )
             return str(netaddr.IPAddress(int(v.ip) ^ 3))
-        raise Exception("Not a point-to-point network")
+        raise JinjaFx.TemplateError("Not a point-to-point network")
 
 
 def _prefix_query(v):
@@ -290,7 +295,7 @@ def _prefix_query(v):
 def _previous_usable_query(v, vtype):
     if vtype == "address":
         # Does it make sense to raise an error
-        raise Exception("Not a network address")
+        raise JinjaFx.TemplateError("Not a network address")
     elif vtype == "network":
         if v.size > 1:
             first_usable, last_usable = _first_last(v)
@@ -321,7 +326,7 @@ def _public_query(v, value):
 def _range_usable_query(v, vtype):
     if vtype == "address":
         # Does it make sense to raise an error
-        raise Exception("Not a network address")
+        raise JinjaFx.TemplateError("Not a network address")
     elif vtype == "network":
         if v.size > 1:
             first_usable, last_usable = _first_last(v)
@@ -405,6 +410,9 @@ def _empty_hwaddr_query(v, value):
         return value
 
 
+class mac_linux(netaddr.mac_unix):
+  word_fmt = "%.2x"
+
 def _linux_query(v):
     v.dialect = mac_linux
     return str(v)
@@ -431,7 +439,7 @@ def _win_query(v):
 # the inputs.
 def cidr_merge(value, action="merge"):
     if not hasattr(value, "__iter__"):
-        raise Exception(
+        raise JinjaFx.TemplateError(
             "cidr_merge: expected iterable, got " + repr(value)
         )
 
@@ -439,9 +447,9 @@ def cidr_merge(value, action="merge"):
         try:
             return [str(ip) for ip in netaddr.cidr_merge(value)]
         except Exception as e:
-            raise Exception(
+            raise JinjaFx.TemplateError(
                 "cidr_merge: error in netaddr:\n%s" % e
-            )
+            ) from None
 
     elif action == "span":
         # spanning_cidr needs at least two values
@@ -451,19 +459,19 @@ def cidr_merge(value, action="merge"):
             try:
                 return str(netaddr.IPNetwork(value[0]))
             except Exception as e:
-                raise Exception(
+                raise JinjaFx.TemplateError(
                     "cidr_merge: error in netaddr:\n%s" % e
-                )
+                ) from None
         else:
             try:
                 return str(netaddr.spanning_cidr(value))
             except Exception as e:
-                raise Exception(
+                raise JinjaFx.TemplateError(
                     "cidr_merge: error in netaddr:\n%s" % e
-                )
+                ) from None
 
     else:
-        raise Exception(
+        raise JinjaFx.TemplateError(
             "cidr_merge: invalid action '%s'" % action
         )
 
@@ -494,7 +502,7 @@ def ipaddr(value, query="", version=False, alias="ipaddr"):
         "wrap": ("vtype", "value"),
     }
 
-    query_func_map = {
+    query_func_map: Dict[str, Callable] = {
         "": _empty_ipaddr_query,
         "6to4": _6to4_query,
         "address": _ip_query,
@@ -687,9 +695,9 @@ def ipaddr(value, query="", version=False, alias="ipaddr"):
                 return value
 
         except Exception:
-            raise Exception(
+            raise JinjaFx.TemplateError(
                 alias + ": unknown filter type: %s" % query
-            )
+            ) from None
 
     return False
 
@@ -702,14 +710,14 @@ def ipmath(value, amount):
             ip = netaddr.IPAddress(value)
     except (netaddr.AddrFormatError, ValueError):
         msg = "You must pass a valid IP address; {0} is invalid".format(value)
-        raise Exception(msg)
+        raise JinjaFx.TemplateError(msg) from None
 
     if not isinstance(amount, int):
         msg = (
             "You must pass an integer for arithmetic; "
             "{0} is not a valid integer"
         ).format(amount)
-        raise Exception(msg)
+        raise JinjaFx.TemplateError(msg)
 
     return str(ip + amount)
 
@@ -825,13 +833,13 @@ def ipsubnet(value, query="", index="x"):
             msg = "You must pass a valid subnet or IP address; {0} is invalid".format(
                 query_string
             )
-            raise Exception(msg)
+            raise JinjaFx.TemplateError(msg)
         query = netaddr.IPNetwork(v)
         for i, subnet in enumerate(query.subnet(value.prefixlen), 1):
             if subnet == value:
                 return str(i)
         msg = "{0} is not in the subnet {1}".format(value.cidr, query.cidr)
-        raise Exception(msg)
+        raise JinjaFx.TemplateError(msg)
     return False
 
 
@@ -881,7 +889,7 @@ def next_nth_usable(value, offset):
         return False
 
     if type(offset) != int:
-        raise Exception("Must pass in an integer")
+        raise JinjaFx.TemplateError("Must pass in an integer")
     if v.size > 1:
         first_usable, last_usable = _first_last(v)
         nth_ip = int(netaddr.IPAddress(int(v.ip) + offset))
@@ -903,7 +911,7 @@ def previous_nth_usable(value, offset):
         return False
 
     if type(offset) != int:
-        raise Exception("Must pass in an integer")
+        raise JinjaFx.TemplateError("Must pass in an integer")
     if v.size > 1:
         first_usable, last_usable = _first_last(v)
         nth_ip = int(netaddr.IPAddress(int(v.ip) - offset))
@@ -1066,7 +1074,7 @@ def hwaddr(value, query="", alias="hwaddr"):
 
     query_func_extra_args = {"": ("value",)}
 
-    query_func_map = {
+    query_func_map: Dict[str, Callable] = {
         "": _empty_hwaddr_query,
         "bare": _bare_query,
         "bool": _bool_hwaddr_query,
@@ -1086,9 +1094,9 @@ def hwaddr(value, query="", alias="hwaddr"):
     except Exception:
         v = None
         if query and query != "bool":
-            raise Exception(
+            raise JinjaFx.TemplateError(
                 alias + ": not a hardware address: %s" % value
-            )
+            ) from None
 
     extras = []
     for arg in query_func_extra_args.get(query, tuple()):
@@ -1096,9 +1104,9 @@ def hwaddr(value, query="", alias="hwaddr"):
     try:
         return query_func_map[query](v, *extras)
     except KeyError:
-        raise Exception(
+        raise JinjaFx.TemplateError(
             alias + ": unknown filter type: %s" % query
-        )
+        ) from None
 
 
 def macaddr(value, query=""):
@@ -1106,7 +1114,7 @@ def macaddr(value, query=""):
 
 
 def _need_netaddr(f_name, *args, **kwargs):
-    raise Exception(
+    raise JinjaFx.TemplateError(
         "The %s filter requires python's netaddr be "
         "installed on the ansible controller" % f_name
     )
